#include "butterworth.h"

volatile int pa1,pa2,pa3, sa[400];
volatile int16_t uf,oen,init_done;
volatile float pwr[3], ac_output[3],pot_pwr;
volatile int16_t ipwr[3];
volatile int sst;
 
float Input[3], Output[3], Setpoint[3];
float errSum[3], lastErr[3], kp, ki, kd;

void check_switch_cont(void);
void do_pid(int);
void init_vars(void);

HardwareTimer t1(TIM1);

volatile uint8_t blcd,go_lcd = 0,bpos=0;
volatile int baudc = 0, baud_waits=1;
volatile long lcd_wait=0;
char buf[85],*bp,bcount;         // local text buffer for LCD


// adc stuff
volatile uint32_t *adcsr =         (uint32_t *)(0x40012000);
volatile uint32_t *adccr1 =       (uint32_t *)(0x40012004);
volatile uint32_t *adccr2 =       (uint32_t *)(0x40012008);
volatile uint32_t *adcsmpr1 =       (uint32_t *)(0x4001200c);
volatile uint32_t *adcsmpr2 =       (uint32_t *)(0x40012010);
volatile uint32_t *adcdr =        (uint32_t *)(0x4001204c);
volatile uint32_t *adcsqr1 =        (uint32_t *)(0x4001202c);
volatile uint32_t *adcsqr3 =        (uint32_t *)(0x40012034);
volatile uint32_t *adcccr =         (uint32_t *)(0x40012304);
volatile uint32_t *rcccfgr =      (uint32_t *)(0x40023808);

void setupadc()
{
  // do this last, after pwm pin setup
  GPIO_InitTypeDef GPIO_InitDef;
   __HAL_RCC_GPIOA_CLK_ENABLE();
  GPIO_InitDef.Pin = GPIO_PIN_0 |GPIO_PIN_1 |GPIO_PIN_2 |GPIO_PIN_3  ;      
  GPIO_InitDef.Mode = GPIO_MODE_ANALOG;
  GPIO_InitDef.Pull = GPIO_NOPULL;
  HAL_GPIO_Init(GPIOA, &GPIO_InitDef);
  __HAL_RCC_ADC1_CLK_ENABLE();
  *adccr1 = 0x00000000;   // 0x0n000000 where n=0,1,2,3 for 12,10,8 and 6 bit rez ADC
  *adccr2 |= 1 ;    // ADC ON
  *adcccr =  ADC_TWOSAMPLINGDELAY_5CYCLES;
  *adcsqr1 = 0;
  *adcsmpr2 = 0;
  *adcsmpr1 = 0;
  *adcsqr3 = 1;     
}

int get_adc(int c)
{
*adcsqr3 = c;
*adccr2 |= 0x40000000;
while(!(*adcsr & 2))
  ;
return  *adcdr;
}




void do_pid(int i)
{   
   float error, dErr, timeChange = 0.01;    
   error = Setpoint[i] - Input[i];
   errSum[i] += (error * timeChange);
   dErr = (error - lastErr[i]) / timeChange;
   Output[i] = kp * error + ki * errSum[i] + kd * dErr;
   lastErr[i] = error;
}

void init_vars()
  {
  int i;
  kp = 0.1;     // set up PID for AC output voltage control
  ki = 0.0;
  kd = 0.01;
  errSum[0] = errSum[1] = errSum[2] = 0.0;       
  lastErr[0] = lastErr[1] = lastErr[2] = 0.0;
  Setpoint[0] = Setpoint[1] = Setpoint[2] = 0.75;
  sst = 0;
  pwr[0] = pwr[1] = pwr[2] = 0.0;
  ipwr[0] = ipwr[1] = ipwr[2] = 0;
  uf = 0;
  ac_output[0] = ac_output[1] = ac_output[2] = 0.0;
  bcount = 0;
  bp = &buf[0];
  for(i=0; i < 80; i++)
    buf[i] = ' ';
  pot_pwr = 0.0;
  }

void lcd_bit()
  {

   
// send one char out pin. First, get a character and put it in "blcd" and init state machine.
if (go_lcd == 0 && baudc > 20)    // not send as soon as possible, let the LCD digest the data with a delay here
  {
  // get next char
  blcd = buf[bcount];
  bcount++;
  if (bcount > 80)
    {
      GPIOC->BSRR = 0x4000;
    bcount = 0;
    blcd =  '$';     // this is translated by the serial LCD as "put cursor at top left position"
    GPIOC->BSRR = 0x40000000;
    }   
  go_lcd = 1;
  bpos = 0;
  baudc = 2;
  }
  
if (go_lcd == 1 && baudc > baud_waits)   // a little 8 bit serial output state machine
  {
    if(bpos == 9)
    {
    // send stop bit, pull high
    GPIOB->BSRR = 0x0400;
    go_lcd = 0;
    baudc = 0;
    bpos = 0;
    }
  if (bpos > 0  )
    {
      if (blcd & 1)
        GPIOB->BSRR = 0x0400;
      else
        GPIOB->BSRR = 0x04000000;
      blcd = blcd >> 1;
      bpos++;  
      baudc = 0;
    }
  if(bpos == 0 && go_lcd == 1)
    {
    // send start bit, pull low
    GPIOB->BSRR = 0x04000000;
    bpos++;  
    baudc = 0;
    }
  }
  baudc++;  
  }

void t1_int(void)
{
  if ((TIM1->SR & 0x0010) == 0x0010)
   {  
     TIM1->SR =0;    // reset int flag - mandatory or else it fires continuously
    pa1++;   //clock the phase1 angle counter
    if (pa1 == 199)
      {
        uf = 1;
        GPIOC->BSRR = 0x8000;    //send a sync pulse for DSO, square wave
      }
    if (pa1 >= 400)    // done 360 degrees?
      {
        pa1 = 0;       // then restart at 0 degrees
        GPIOC->BSRR = 0x80000000;    //send a sync pulse for DSO
        uf=1;
      }   
    pa2++; if (pa2 >= 400) pa2 = 0;
    pa3++; if (pa3 >= 400) pa3 = 0;  
    TIM1->CCR1 = (sa[pa1] * ipwr[0]) >> 8;     // phase 1 pwm width
    TIM1->CCR2 = (sa[pa2] * ipwr[0]) >> 8;     // phase 2
    TIM1->CCR3 = (sa[pa3] * ipwr[0]) >> 8;     // phase 3
   do_ctl_loop();
   lcd_bit();
   }
}

void do_ctl_loop()
  {
  int i;
  float incpwr = 1.0/255.0;
  if (uf == 1)
  {  
    uf = 0;
    for(i=0; i < 3; i++)                         
      { 
      /*  
      if(ac_output[i] < pot_pwr)
        pwr[i] += incpwr;     
      else                    
        pwr[i] -= incpwr;           
      ipwr[i] = (int)( (float)(256.0) * pwr[i] );   
      */
         
      ipwr[i] = (int)( (float)(256.0) * pot_pwr );   
               
      }
  check_switch_cont();    
  }
}


void check_switch_cont()
{   
  if ((GPIOB->IDR & 0x0200) == 0)     // pin grounded
    {
    if (init_done == 0 && sst == 0)
      {
      init_vars();
      init_done = 1;
      }  
    oen = 1;      // slow start
    GPIOC->BSRR = 0x20000000;     // PC13 LED inverted logic, this resets bit and lights LED
    }
else
  {
    oen = 0;    
    if (sst > 0)
      init_done = 0;
    GPIOC->BSRR = 0x00002000;   // sets bit so dims LED
  }
}


void setup() 
{
int i;
float th,p;
delay(5000);
// LP filter setup  Fc,N_order,filter_index
get_biquad(0.1,4,0);    
get_biquad(0.1,4,1);   
get_biquad(0.1,4,2);   
get_biquad(0.1,4,3); 
 
// sine() lookup table, 
for (i=0; i < 400; i++)
  {
    th = 2.0 * 3.14159 * (float)i / 400.0;
    p = (1.0 + sin(th)) * 2100;
    sa[i] = (int)p;
  }

init_vars();  
init_done = 1;
  
pa1 = 0;         // init phase angle = zero
pa2 = 133;        //  these 3 get incremented each time in interrupt code
pa3 = 266;        // at 20KHz. 399 elements in lookup table at 20KHz = 50Hz

// TIMER1 configs
PWM_pin_config();
Timer1_config();
t1.attachInterrupt(t1_int);
HAL_NVIC_SetPriority(TIM1_UP_TIM10_IRQn ,1,3);
TIM1->DIER = 0x0001;    // let it rip, enable TIMER1

// PC13 LED setup and PC15 for 50Hz sync pulse
pinMode(PC13,OUTPUT);         // LED
pinMode(PC14,OUTPUT);         // debug
pinMode(PC15,OUTPUT);         // DSO sync 50Hz
pinMode(PB9,INPUT_PULLUP);    // inverter run, pull down to ground for go
pinMode(PB10,OUTPUT);         // LCD 9600 baud TX

setupadc();

}

void loop() 
{
int i;
float v;
char fbuf[20],gbuf[20],hbuf[20];
float minpwr = 0.01, maxpwr = 0.99;
// approx sample rate for the 3 inputs is 5.4KHz with STM32F401 at 84MHz using analogRead()
// approx sample rate for the 3 inputs is 85KHz with STM32F401 at 84MHz using get_adc()

for(i=0; i < 3; i++)
  {
    //v = (float)analogRead(i) / (float)(1024.0);  
    v = get_adc(i) / (float)(4.0 * 1024.0);
    biquad_filter(&v,i);
    ac_output[i] = v;
    //ac_output[i] = 0.0;         // for testing
  }
  
// get pot setting, filter
v = get_adc(3) / (float)(4.0 * 1024.0);
biquad_filter(&v,3);
if (v > maxpwr) v = maxpwr;     
if (v < minpwr) v = minpwr;  
pot_pwr = v;
      
lcd_wait++;
if (lcd_wait > 6100)
  {
    lcd_wait=0;
    dtostrf(pot_pwr,4,3,fbuf);
    //dtostrf(pwr[1],4,3,gbuf);
    //dtostrf(pwr[2],4,3,hbuf);
    sprintf(bp,"Pot -> %5s     ",fbuf); 
    dtostrf(ac_output[0],4,2,fbuf);
    dtostrf(ac_output[1],4,2,gbuf);
    dtostrf(ac_output[2],4,2,hbuf);
    sprintf(bp+40,"O %4s %4s %4s",fbuf,gbuf,hbuf); 
    dtostrf(pwr[0],4,2,fbuf);
    dtostrf(pwr[i],4,2,gbuf);
    dtostrf(pwr[2],4,2,hbuf);
    sprintf(bp+20,"P %4s %4s %4s",fbuf,gbuf,hbuf); 
  }
 
}



void Timer1_config(void)
{
  //
  // timer1, ch1,2,3 OC and OCN outputs enabled with deadtime
  // correct signaling for 3 phase inverter
  //
  uint16_t ccmr1=0, ccmr2=0,  ccer=0;
  __HAL_RCC_TIM1_CLK_ENABLE();
  TIM1->BDTR = 0x0c50;    // DT part is low 8 bits
  TIM1->PSC = 0;
  TIM1->ARR = 4200;
  TIM1->CR1 = 0;
  TIM1->RCR = 0;

  // ch1,2 TIM_OCPreload_Enable & PWM mode 1
  // ch1,2,3 OC and OCN config   OC enabled, active high, OCN enabled, active high
  TIM1->CCER = 0x0555;
  TIM1->CCMR1 = 0x6868;
  TIM1->CCMR2 = 0x0068;

  TIM1->CCR1 = 1;     // ch1 pwm width
  TIM1->CCR2 = 1;     // ch2 pwm width
  TIM1->CCR3 = 1;     // ch3 pwm width
  TIM1->CCR4 = 1;     
  
  TIM1->CR2 = 0;
  TIM1->CR1 |= 0x0080;    // enable preload
  TIM1->BDTR |= 0x8000;   // enable PWM outputs, needed for timer1
  TIM1->CR1 |= 0x0001;    // enable timer
}


void PWM_pin_config(void)
  {
  // manual GPIO setup, the HAL functions do not work with TIMER1 ALT1 complimentary output pins
  // timer 1, ch 1,2,3 alt functions out A8,9,10  
  __HAL_RCC_GPIOA_CLK_ENABLE();
  GPIOA->PUPDR =   0x00154000;    // 10,9,8 pins pull up  01
  GPIOA->OSPEEDR = 0x003fC000;    // speed is 11 or fastest
  GPIOA->MODER =   0x002a8000;    // Alt function
  GPIOA->OTYPER = 0;              // output push-pull
  GPIOA->AFR[1] = 0x00000111;     // AF = 0001 = AF1
  // timer1 ch 1,2,3 complimentary outputs B13,14,15
  __HAL_RCC_GPIOB_CLK_ENABLE();
  GPIOB->PUPDR =   0x54000000;
  GPIOB->OSPEEDR = 0xfc000000;
  GPIOB->MODER =   0xa8000000;
  GPIOB->OTYPER = 0;
  GPIOB->AFR[1] =  0x11100000;
  }
